/*
 * Copyright (c) Citrix Systems, Inc
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>

#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <openssl/bio.h>
#include <openssl/evp.h>

#include <debug.h>
#include <efi.h>
#include <serialize.h>
#include <oxtdb.h>
#include <xapidb.h>

// db over dbus
#include <xcdbus.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <event.h>
#include "rpcgen/db_client.h"

#define OXTDB_DBUS_SERVICE "com.citrix.xenclient.db"
#define OXTDB_DBUS_PATH    "/"

#define MAX_OXTDB_PATH_SIZE 512
#define MAX_OXTDB_VALUE_SIZE 16

static xcdbus_conn_t * xcdbus_conn = NULL;

/* The VM's uuid. Used for saving to the db. */
char *oxtdb_arg_uuid;

// DoS protection
#define MAX_CREDIT        100
#define CREDIT_PER_SECOND 2
#define NS_PER_CREDIT (1000000000 / CREDIT_PER_SECOND)
static time_t last_time; /* Time of the last send. */
static unsigned int send_credit = MAX_CREDIT; /* Number of allowed fast sends. */

static void
maybe_sleep(void)
{
    time_t cur_time, diff_time;

    /*
     * To avoid a DoS by the VM, rate limit writes.
     * Normal usage should never hit this.
     */
    cur_time = time(NULL);
    diff_time = cur_time - last_time;
    last_time = cur_time;
    send_credit += diff_time * CREDIT_PER_SECOND;
    if (send_credit > MAX_CREDIT)
        send_credit = MAX_CREDIT;

    if (send_credit > 0) {
        send_credit--;
    } else {
        /* If no credit, wait the correct amount of time to get a credit. */
        struct timespec ts = {0, NS_PER_CREDIT};

        nanosleep(&ts, NULL);
        last_time = time(NULL);
    }
}

static int oxtdb_dbus_init(void) {
    GError * err = NULL;
    DBusGConnection * gdbus_conn;

    if (xcdbus_conn) {
        return 0;
    }

    event_init();
    g_type_init();

    gdbus_conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &err);

    if (!gdbus_conn) {
        g_error_free(err);
        ERR("Unable to initialize dbus connection\n");
        return -1;
    }

    xcdbus_conn = xcdbus_init_event("com.citrix.xenclient.db", gdbus_conn);
    if (!xcdbus_conn) {
        ERR("Unable to initialize xcdbus service\n");
        return -1;
    }

    return 0;
}

// Write a value to the specified DB path.
static bool oxtdb_write(char * value, const char * path_format, ...) {

    va_list args;
    char * path = malloc(MAX_OXTDB_PATH_SIZE);

    if (!path) {
        ERR("Out of memory!\n");
        goto err;
    }

    if (oxtdb_dbus_init() == -1 || !xcdbus_conn) {
        goto err;
    }

    va_start(args, path_format);
    vsnprintf(path, MAX_OXTDB_PATH_SIZE, path_format, args);
    va_end(args);

    if (!com_citrix_xenclient_db_write_(xcdbus_conn, OXTDB_DBUS_SERVICE, OXTDB_DBUS_PATH, path, value)) {
        ERR("db_write op failed. tried writing %s to %s\n", value, path);
        goto err;
    }

    // success
    free(path);
    return true;

err:
    if (path)
        free(path);
    return false;
}

// Determine if db node exists.
static bool oxtdb_exists(const char * path_format, ...) {

    int found = false;

    va_list args;
    char * path = malloc(MAX_OXTDB_PATH_SIZE);

    if (!path) {
        ERR("Out of memory!\n");
        goto err;
    }

    if (oxtdb_dbus_init() == -1 || !xcdbus_conn) {
        goto err;
    }

    va_start(args, path_format);
    vsnprintf(path, MAX_OXTDB_PATH_SIZE, path_format, args);
    va_end(args);

    if (!com_citrix_xenclient_db_exists_(xcdbus_conn, OXTDB_DBUS_SERVICE, OXTDB_DBUS_PATH, path, &found)) {
        ERR("db_exists op failed. path attempted: %s\n", path);
        goto err;
    }

    // success
    free(path);
    return (bool)found;

err:
    if (path)
        free(path);
    return false;
}

// Allocates memory!
// Gets the value of the specified DB key.
// The string returned should be freed.
static char * oxtdb_read(const char * path_format, ...) {
    char * string;
    va_list args;
    char * path = malloc(MAX_OXTDB_PATH_SIZE);

    if (!path) {
        ERR("Out of memory!\n");
        goto err;
    }

    if (oxtdb_dbus_init() == -1 || !xcdbus_conn) {
        goto err;
    }

    va_start(args, path_format);
    vsnprintf(path, MAX_OXTDB_PATH_SIZE, path_format, args);
    va_end(args);

    if (!com_citrix_xenclient_db_read_(xcdbus_conn, OXTDB_DBUS_SERVICE, OXTDB_DBUS_PATH, path, &string)) {
        ERR("db_read op failed. path attempted: %s\n", path);
        goto err;
    }

    // success
    free(path);
    return string;

err:
    if (path)
        free(path);
    return NULL;
}

size_t
oxtdb_serialize_variables(uint8_t **out, size_t *len, bool only_nv)
{
    return xapidb_serialize_variables(out, len, only_nv);
}

static bool
write_all(const char *buf)
{
    bool check = false;
    check = oxtdb_exists("/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid);
    if (check) {
        char *temp = oxtdb_read("/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid);
        if (!temp) {
            ERR("unable to read oxtdb");
            return false;
        }
        if (strcmp(buf, temp) == 0) {
            free(temp);
            ERR("found match in write");
            return false;
        }
        if (temp != NULL) {
            free(temp);
        }
    }
    return oxtdb_write((char *)buf, "/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid);
}

static void
read_all(char **buf)
{
    bool ret = false;

    ret = oxtdb_exists("/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid);
    if (ret) {
        DBG("attempting to cat %d\n");
        *buf = oxtdb_read("/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid);
    } else {
        *buf = NULL;
    }
}

bool
oxtdb_set_variable(void)
{
    uint8_t *buf;
    char *encoded;
    size_t len;
    bool ret = false;

    if (!oxtdb_arg_uuid)
        return true;

    oxtdb_serialize_variables(&buf, &len, true);
    if (len == 0)
        return false;

    if (!base64_encode(buf, len, &encoded)) {
        free(buf);
        return false;
    }
    free(buf);

    // DoS protection
    maybe_sleep();

    ret = write_all(encoded);
    free(encoded);

    return ret;
}

bool
oxtdb_parse_blob(uint8_t **buf, int len) {
    return xapidb_parse_blob(buf, len);
}

enum backend_init_status
oxtdb_init(void)
{
    char *encoded = NULL;
    uint8_t *buf, *ptr = NULL;
    BIO *bio, *b64 = NULL;
    int max_len, n, total = 0;
    int ret = 0;

    if (!oxtdb_exists("/vm/%s/varstored-efivar-blob/0", oxtdb_arg_uuid)) {
        return BACKEND_INIT_FIRSTBOOT;
    }

    read_all(&encoded);

    if (!encoded) {
        goto err;
    }

    max_len = strlen(encoded) * 3 / 4;

    buf = malloc(max_len);
    if (!buf) {
        ERR("Failed to allocate memory\n");
        goto err;
    }

    bio = BIO_new_mem_buf(encoded, -1);
    if (!bio) {
        ERR("Failed to create BIO\n");
        goto err;
    }
    b64 = BIO_new(BIO_f_base64());
    if (!b64) {
        ERR("Failed to create BIO\n");
        goto err;
    }
    BIO_push(b64, bio);
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);

    for (;;) {
        n = BIO_read(b64, buf + total, max_len - total);
        if (n <= 0)
            break;
        total += n;
    }

    ptr = buf;
    DBG("total: %d\n", total);
    ret = oxtdb_parse_blob(&ptr, total);

    if (!ret) {
        ERR("oxtdb parse operation failed\n");
        goto err;
    }

    // success
    BIO_free_all(b64);
    free(buf);
    free(encoded);

    return BACKEND_INIT_SUCCESS;

err:
    if (bio) {
        if (b64)
            BIO_free_all(b64);
        else
            BIO_free_all(bio);
    }
    if (buf)
        free(buf);
    if (encoded)
        free(encoded);
    return BACKEND_INIT_FAILURE;
}

bool
oxtdb_sb_notify(void)
{
    ERR("The VM failed to pass Secure Boot verification.");
    return true;
}
